﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Configuration;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Web;
using System.Web.Mvc;
using ThinkGeo.MapSuite.Core;
using ThinkGeo.MapSuite.MvcEdition;

namespace StyleBug.Controllers
{    public class HomeController : Controller
    {
        private const string GAINFUL_AREAS = "GAINFUL_AREAS";
        private const string SALARY = "SALARY";
        private const string AREA_ID = "AREA_ID";
        private const string AREA_DESCRIPTION = "AREA_DESCRIPTION";
        private const string STATE_LAYER = "STATE_LAYER";

        private static int count = 0;

        public enum GainfulShapeType
        {
            State = 1,
            StatisticalArea = 2
        }

        public class MapViewModel
        {
            public Map ViewMap { get; set; }
            public string Memory { get; set; }

            public int Count { get; set; }

            public int KeyCount { get; set; }
        }

        public class GainfulAreaShapeModel
        {
            public int AreaId { get; set; }
            public string AreaDescription { get; set; }
            public int ShapeId { get; set; }
            public string ShapeDescription { get; set; }
            public GainfulShapeType ShapeType { get; set; }
        }

        public interface IGainfulSalary
        {
            decimal? Salary { get; set; }
        }

        public class GainfulAreaSalaryModel : GainfulAreaShapeModel, IGainfulSalary
        {
            public decimal? Salary { get; set; }
        }

        //
        // GET: /Home/

        public ActionResult Index()
        {
            count = 0;
            return View();
        }

        public ActionResult MapPartial()
        {
            // Create the map
            Map map = new Map("Map", 350, 350);
            map.MapBackground.BackgroundBrush = new GeoSolidBrush(GeoColor.FromHtml("#E5E3DF"));
            map.CurrentExtent = new RectangleShape(-125.0011, 49.5904, -66.9326, 24.9493);
            map.MapUnit = GeographyUnit.DecimalDegree;
            map.WebConfigRegisterMode = WebConfigRegistrationMode.Manual;

            LayerOverlay overlay = new LayerOverlay();
            overlay.IsBaseOverlay = true;

            string connectionString = @"Data Source = D:\downloads\heatmap\WorldMapKit-4326-20151011-641421321.sqlite;Version=3;Read Only=True;";
            OsmWorldMapKitLayer osmWorldMapKitLayer = new OsmWorldMapKitLayer(connectionString, OsmWorldMapKitDatabaseType.Sqlite);
            overlay.Layers.Add(osmWorldMapKitLayer);

            map.CustomOverlays.Add(overlay);

            bool exhibitLeak = true;
            if (exhibitLeak)
            {
                LayerOverlay shapeOverlay = CreateGainfulAreaOverlay();
                shapeOverlay.TileType = TileType.SingleTile;

                // Create multiple layers just to make memory use more obvious
                for (int k = 0; k < 10; k++)
                {
                    ColorizeMap(shapeOverlay, "05", "12.0101", k);
                }

                map.CustomOverlays.Add(shapeOverlay);
            }

            long memory; //= GC.GetTotalMemory(true);

            using (Process process = Process.GetCurrentProcess())
            {
                process.Refresh();
                memory = process.PrivateMemorySize64;
            }

            MapViewModel model = new MapViewModel();
            model.ViewMap = map;
            model.Memory = string.Format("{0:F1} MB", (double)memory / 1024 / 1024);

            model.Count = ++count;

            model.KeyCount = Session.Keys.Count;

            return View(model);
        }

        private LayerOverlay CreateGainfulAreaOverlay()
        {
            LayerOverlay layerOverlay = new LayerOverlay(GAINFUL_AREAS);
            layerOverlay.IsBaseOverlay = false;
            string tileType = ConfigurationManager.AppSettings["GainfulTileType"];
            if (!string.IsNullOrWhiteSpace(tileType) && tileType == "single")
            {
                layerOverlay.TileType = TileType.SingleTile;
            }
            else
            {
                layerOverlay.TileType = TileType.MultipleTile;
            }
            return layerOverlay;
        }

        private void ColorizeMap(LayerOverlay layerOverlay, string credentialLevel, string cipCode, int count)
        {
            IList<GainfulAreaSalaryModel> salaries = new List<GainfulAreaSalaryModel>() {
                  new GainfulAreaSalaryModel() {
                    Salary = 86793.5000M,
                    AreaId = 49,
                    AreaDescription = "Utah",
                    ShapeId = 44,
                    ShapeDescription = "Utah",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 72811.0000M,
                    AreaId = 17,
                    AreaDescription = "Illinois",
                    ShapeId = 13,
                    ShapeDescription = "Illinois",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 71383.0000M,
                    AreaId = 55,
                    AreaDescription = "Wisconsin",
                    ShapeId = 49,
                    ShapeDescription = "Wisconsin",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 70890.0000M,
                    AreaId = 36,
                    AreaDescription = "New York",
                    ShapeId = 32,
                    ShapeDescription = "New York",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 70320.5000M,
                    AreaId = 13,
                    AreaDescription = "Georgia",
                    ShapeId = 10,
                    ShapeDescription = "Georgia",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 70125.0000M,
                    AreaId = 5,
                    AreaDescription = "Arkansas",
                    ShapeId = 3,
                    ShapeDescription = "Arkansas",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 64345.0000M,
                    AreaId = 11,
                    AreaDescription = "District of Columbia",
                    ShapeId = 8,
                    ShapeDescription = "District of Columbia",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 64328.0000M,
                    AreaId = 26,
                    AreaDescription = "Michigan",
                    ShapeId = 22,
                    ShapeDescription = "Michigan",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 63665.0000M,
                    AreaId = 9,
                    AreaDescription = "Connecticut",
                    ShapeId = 6,
                    ShapeDescription = "Connecticut",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 63478.0000M,
                    AreaId = 24,
                    AreaDescription = "Maryland",
                    ShapeId = 20,
                    ShapeDescription = "Maryland",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 62585.5000M,
                    AreaId = 6,
                    AreaDescription = "California",
                    ShapeId = 4,
                    ShapeDescription = "California",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 61055.5000M,
                    AreaId = 20,
                    AreaDescription = "Kansas",
                    ShapeId = 16,
                    ShapeDescription = "Kansas",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 59219.5000M,
                    AreaId = 19,
                    AreaDescription = "Iowa",
                    ShapeId = 15,
                    ShapeDescription = "Iowa",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 59117.5000M,
                    AreaId = 37,
                    AreaDescription = "North Carolina",
                    ShapeId = 33,
                    ShapeDescription = "North Carolina",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 57094.5000M,
                    AreaId = 53,
                    AreaDescription = "Washington",
                    ShapeId = 47,
                    ShapeDescription = "Washington",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 56975.5000M,
                    AreaId = 18,
                    AreaDescription = "Indiana",
                    ShapeId = 14,
                    ShapeDescription = "Indiana",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 56856.5000M,
                    AreaId = 33,
                    AreaDescription = "New Hampshire",
                    ShapeId = 29,
                    ShapeDescription = "New Hampshire",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 55709.0000M,
                    AreaId = 12,
                    AreaDescription = "Florida",
                    ShapeId = 9,
                    ShapeDescription = "Florida",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 53941.0000M,
                    AreaId = 2,
                    AreaDescription = "Alaska",
                    ShapeId = 1,
                    ShapeDescription = "Alaska",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 53490.5000M,
                    AreaId = 41,
                    AreaDescription = "Oregon",
                    ShapeId = 37,
                    ShapeDescription = "Oregon",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 52963.5000M,
                    AreaId = 27,
                    AreaDescription = "Minnesota",
                    ShapeId = 23,
                    ShapeDescription = "Minnesota",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 52921.0000M,
                    AreaId = 56,
                    AreaDescription = "Wyoming",
                    ShapeId = 50,
                    ShapeDescription = "Wyoming",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 52122.0000M,
                    AreaId = 35,
                    AreaDescription = "New Mexico",
                    ShapeId = 31,
                    ShapeDescription = "New Mexico",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 51748.0000M,
                    AreaId = 4,
                    AreaDescription = "Arizona",
                    ShapeId = 2,
                    ShapeDescription = "Arizona",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 51714.0000M,
                    AreaId = 51,
                    AreaDescription = "Virginia",
                    ShapeId = 46,
                    ShapeDescription = "Virginia",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 49410.5000M,
                    AreaId = 50,
                    AreaDescription = "Vermont",
                    ShapeId = 45,
                    ShapeDescription = "Vermont",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 48926.0000M,
                    AreaId = 47,
                    AreaDescription = "Tennessee",
                    ShapeId = 42,
                    ShapeDescription = "Tennessee",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 48730.5000M,
                    AreaId = 22,
                    AreaDescription = "Louisiana",
                    ShapeId = 18,
                    ShapeDescription = "Louisiana",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 48467.0000M,
                    AreaId = 48,
                    AreaDescription = "Texas",
                    ShapeId = 43,
                    ShapeDescription = "Texas",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 48254.5000M,
                    AreaId = 39,
                    AreaDescription = "Ohio",
                    ShapeId = 35,
                    ShapeDescription = "Ohio",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 46138.0000M,
                    AreaId = 30,
                    AreaDescription = "Montana",
                    ShapeId = 26,
                    ShapeDescription = "Montana",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 45109.5000M,
                    AreaId = 72,
                    AreaDescription = "Puerto Rico",
                    ShapeId = 54,
                    ShapeDescription = "Puerto Rico",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 44956.5000M,
                    AreaId = 8,
                    AreaDescription = "Colorado",
                    ShapeId = 5,
                    ShapeDescription = "Colorado",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 44939.5000M,
                    AreaId = 10,
                    AreaDescription = "Delaware",
                    ShapeId = 7,
                    ShapeDescription = "Delaware",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 44557.0000M,
                    AreaId = 45,
                    AreaDescription = "South Carolina",
                    ShapeId = 40,
                    ShapeDescription = "South Carolina",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 43724.0000M,
                    AreaId = 38,
                    AreaDescription = "North Dakota",
                    ShapeId = 34,
                    ShapeDescription = "North Dakota",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 41378.0000M,
                    AreaId = 31,
                    AreaDescription = "Nebraska",
                    ShapeId = 27,
                    ShapeDescription = "Nebraska",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 39584.5000M,
                    AreaId = 40,
                    AreaDescription = "Oklahoma",
                    ShapeId = 36,
                    ShapeDescription = "Oklahoma",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 39074.5000M,
                    AreaId = 32,
                    AreaDescription = "Nevada",
                    ShapeId = 28,
                    ShapeDescription = "Nevada",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 36941.0000M,
                    AreaId = 15,
                    AreaDescription = "Hawaii",
                    ShapeId = 11,
                    ShapeDescription = "Hawaii",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 36176.0000M,
                    AreaId = 54,
                    AreaDescription = "West Virginia",
                    ShapeId = 48,
                    ShapeDescription = "West Virginia",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 31093.0000M,
                    AreaId = 16,
                    AreaDescription = "Idaho",
                    ShapeId = 12,
                    ShapeDescription = "Idaho",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 27710.0000M,
                    AreaId = 29,
                    AreaDescription = "Missouri",
                    ShapeId = 25,
                    ShapeDescription = "Missouri",
                    ShapeType = GainfulShapeType.State
                  },
                  new GainfulAreaSalaryModel() {
                    Salary = 26724.0000M,
                    AreaId = 1,
                    AreaDescription = "Alabama",
                    ShapeId = 0,
                    ShapeDescription = "Alabama",
                    ShapeType = GainfulShapeType.State
                  }
            };

            // Set random salaries
            Random rand = new Random();
            foreach (var sal in salaries)
            {
                sal.Salary = rand.Next(20000, 70000);
            }

            // Remove random features
            for (int k = 0; k < 10; k++)
            {
                int pos = rand.Next(0, salaries.Count);
                salaries.RemoveAt(pos);
            }

            AddCipSalaryFeatures(layerOverlay, salaries, count);
        }

        private void AddCipSalaryFeatures(LayerOverlay layerOverlay, IList<GainfulAreaSalaryModel> salaries, int count)
        {
            // fastClimb causes memory to go up faster, also more likely to get invalid record error
            bool fastClimb = true;
            string shapeFileName = "~/App_Data/USStates.shp";
            InMemoryFeatureLayer stateFeatureLayer = CreateClassBreakInMemoryLayer(Server.MapPath(shapeFileName), salaries, GainfulShapeType.State);

            layerOverlay.Layers.Add(STATE_LAYER + "_" + count, stateFeatureLayer);
        }

        private InMemoryFeatureLayer CreateClassBreakInMemoryLayer(string path, IList<GainfulAreaSalaryModel> salaries, GainfulShapeType shapeType)
        {
            ShapeFileFeatureLayer shapeFileFeatureLayer = new ShapeFileFeatureLayer(path);
            shapeFileFeatureLayer.Open();
            Collection<Feature> allFeatures = shapeFileFeatureLayer.FeatureSource.GetAllFeatures(ReturningColumnsType.AllColumns);
            shapeFileFeatureLayer.Close();

            Collection<Feature> salariedFeatures = AddSalariesToFeatures(allFeatures, salaries);

            List<FeatureSourceColumn> columns = new List<FeatureSourceColumn>()
            {
                new FeatureSourceColumn(SALARY, DbfColumnType.Numeric.ToString(), 10),
                new FeatureSourceColumn(AREA_ID, DbfColumnType.Numeric.ToString(), 10),
                new FeatureSourceColumn(AREA_DESCRIPTION, DbfColumnType.Character.ToString(), 255),
            };

            InMemoryFeatureLayer imf = new InMemoryFeatureLayer(columns, salariedFeatures);

            ClassBreakStyle classBreakStyle = CreateClassBreakStyle(salaries, shapeType);
            imf.ZoomLevelSet.ZoomLevel01.CustomStyles.Add(classBreakStyle);
            imf.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            return imf;
        }

        private Collection<Feature> AddSalariesToFeatures(Collection<Feature> allFeatures, IList<GainfulAreaSalaryModel> featureSalaries)
        {
            // Convert list of salaries to dictionary of feature to index
            var featureToSalary = featureSalaries.Select((s, index) => new { s.ShapeId, index })
                .ToDictionary(pair => pair.ShapeId, pair => pair.index);

            int featureCount = allFeatures.Count;

            Collection<Feature> salariedFeatures = new Collection<Feature>();

            // Add salaries to each feature
            for (int k = 0; k < featureCount; k++)
            {
                // _logger.Debug("Feature " + k + " " + allFeatures[k].ColumnValues["NAME"]);
                if (featureToSalary.ContainsKey(k))
                {
                    int pos = featureToSalary[k];
                    allFeatures[k].ColumnValues.Add(SALARY, featureSalaries[pos].Salary.Value.ToString());
                    allFeatures[k].ColumnValues.Add(AREA_ID, featureSalaries[pos].AreaId.ToString());
                    allFeatures[k].ColumnValues.Add(AREA_DESCRIPTION, featureSalaries[pos].AreaDescription);

                    salariedFeatures.Add(allFeatures[k]);
                }
            }

            return salariedFeatures;
        }

        private ClassBreakStyle CreateClassBreakStyle(IList<GainfulAreaSalaryModel> salaries, GainfulShapeType shapeType)
        {
            GeoColor outlineColor = (shapeType == GainfulShapeType.State ? GeoColor.SimpleColors.Black : GeoColor.SimpleColors.DarkBlue);

            // Assumption is that salaries are ordered by amount, descending
            int yellow = (int)(salaries.Count * 0.3);
            int red = salaries.Count - yellow;

            // Add class break styles based on salaries. Class break style uses values to determine colorization
            // salaries[Count-1] - salaries[red-1]      Are red         bottom 30%
            // salaries[red] - salaries[yellow-1]       Are yellow      middle 40%
            // salaries[yellow] - salaries[0]           Are green       top 30%
            ClassBreakStyle classBreakStyle = new ClassBreakStyle("SALARY");
            // Min value is last salary, so areas with salary less than that won't be colorized
            classBreakStyle.ClassBreaks.Add(new ClassBreak((double)salaries[salaries.Count - 1].Salary.Value, CreateClassBreakAreaStyle(outlineColor, GeoColor.SimpleColors.Red)));
            classBreakStyle.ClassBreaks.Add(new ClassBreak((double)salaries[red].Salary.Value, CreateClassBreakAreaStyle(outlineColor, GeoColor.SimpleColors.Yellow)));
            classBreakStyle.ClassBreaks.Add(new ClassBreak((double)salaries[yellow].Salary.Value, CreateClassBreakAreaStyle(outlineColor, GeoColor.SimpleColors.Green)));

            return classBreakStyle;
        }

        private AreaStyle CreateClassBreakAreaStyle(GeoColor outlineColor, GeoColor brushColor)
        {
            GeoSolidBrush geoSolidBrush = new GeoSolidBrush(new GeoColor(20, brushColor));
            AreaStyle areaStyle = new AreaStyle(new GeoPen(outlineColor), geoSolidBrush);
            return areaStyle;
        }
    }
}
